% function [z] = adimat_logm(x)
%
% Compute z = logm(x). This uses a relatively plain inverse scaling
% and squaring (ISS) method using Pad approximants. The Schur
% decomposition [U T] = schur(x) is extracted from sqrtm and the
% repeated sqrtm is performed on the upper triangular T.
%
% This functions is differentiation with ADiMat to create the runtime
% functions g_adimat_logm, d_adimat_logm, and a_adimat_logm.
%
% see also g_adimat_logm, d_adimat_logm, a_adimat_logm.
%
% This file is part of the ADiMat runtime environment
%
% Copyright 2013 Johannes Willkomm, Institute for Scientific Computing
%                     TU Darmstadt
function [z] = adimat_logm(x)
  n = size(x, 1);
  
  [U T] = adimat_schur(x);
  
  s = 0;
  Ts = T;
  
  nsqrt = 0;
  nsqrtMax = 100;
  while norm(eye(n) - Ts, 1) >= 0.5 && nsqrt < nsqrtMax
    s = s + 1;
    Ts = adimat_sqrtm_triu(Ts);
    nsqrt = nsqrt + 1;
  end
  
  if nsqrt >= nsqrtMax
    warning('adimat:logm:tooManySquareRoots', ...
            'logm: too many square roots (%d)', nsqrt);
  end

%  norm(eye(n) - Ts, 1)
  
  B = pade_logm1(eye(n) - Ts, 8);
%  qual1 = norm(expm(B) - Ts, 1) ./ norm(Ts, 1)

  z = B .* 2.^s;
%  qual2 = norm(expm(z) - T,1) ./ norm(T,1)
  
  z = U * z * U';
%  qual3 = norm(expm(z) - x,1) ./ norm(x,1)
  
end
  
% function [X] = pade_logm1(A, m)
% 
% pade_logm1(I - A, m) computes logm(A) by an [m/m] Pad
% approximant. The result is precise when norm(I - A) is small.
%
% In Higham, "Evaluating pade approximants of the matrix logarithm" it
% is said that partial fraction expansion is the best method, but we
% use a no-frills polynomial evaluation here.
%
function [A] = pade_logm1(A, m)
  n = size(A, 1);

  [p q] = pade_logm1_coeffs(m);
  
  P = polyvalm(fliplr(p), A);
  Q = polyvalm(fliplr(q), A);
  
  % Ai = eye(n);
  % P = zeros(n);
  % Q = zeros(n);
  % for i=1:length(p)
  %   P = P + p(i) .* Ai;
  %   Q = Q + q(i) .* Ai;
  %   Ai = Ai * A;
  % end
  
  A = Q \ P;
end

% function [p q] = pade_logm1_coeffs(m)
% 
% Returns the coefficients of an [m/m] Pad approximant. In reverse
% order for polyvalm, unfortunately.
%
function [p q] = pade_logm1_coeffs(m)
  padeLogm1Coeffs{1} = {
      [0, 2, ]
      [-2, 1, ]
                   };

  padeLogm1Coeffs{2} = {
      [0, 6, -3, ]
      [-6, 6, -1, ]
                   };

  padeLogm1Coeffs{3} = {
      [0, -60, 60, -11, ]
      [60, -90, 36, -3, ]
                   };

  padeLogm1Coeffs{4} = {
      [0, -420, 630, -260, 25, ]
      [420, -840, 540, -120, 6, ]
                   };

  padeLogm1Coeffs{5} = {
      [0, 7560, -15120, 9870, -2310, 137, ]
      [-7560, 18900, -16800, 6300, -900, 30, ]
                   };
  
  padeLogm1Coeffs{6} = {
      [0, 9240, -23100, 20720, -7980, 1218, -49, ]
      [-9240, 27720, -31500, 16800, -4200, 420, -10, ]
                   };
  
  padeLogm1Coeffs{7} = {
      [0, -240240, 720720, -823900, 446600, -115668, 12488, -363, ]
      [240240, -840840, 1164240, -808500, 294000, -52920, 3920, -70, ]
                   };

  padeLogm1Coeffs{8} = {
      [0, -1801800, 6306300, -8768760, 6156150, -2288440, 429660, -34632, 761, ]
      [1801800, -7207200, 11771760, -10090080, 4851000, -1293600, 176400, -10080, 140, ]
                   };

  padeLogm1Coeffs{9} = {
      [0, 61261200, -245044800, 401501100, -346846500, 169279110, -46366320, 6631020, -414810, 7129, ]
      [-61261200, 275675400, -518918400, 529729200, -317837520, 113513400, -23284800, 2494800, -113400, 1260, ]
                   };

  padeLogm1Coeffs{10} = {
      [0, 232792560, -1047566520, 1976694720, -2029787760, 1231242012, -447957510, 94757520, -10694970, 534710, -7381, ]
      [-232792560, 1163962800, -2481078600, 2940537600, -2118916800, 953512560, -264864600, 43243200, -3742200, 138600, -1260, ]
                   };

  padeLogm1Coeffs{11} = {
      [0, -9777287520, 48886437600, -104407463160, 124311227040, -90322696464, 41201832672, -11694132450, 1975433460, -180741990, 7390812, -83711, ]
      [9777287520, -53775081360, 128035908000, -172848475800, 145556611200, -79247488320, 27969701760, -6243237000, 832431600, -59459400, 1829520, -13860, ]
                   };

  padeLogm1Coeffs{12} = {
      [0, -37479602160, 206137811880, -491580289200, 666077712300, -564568516512, 310591833552, -111382663392, 25469343900, -3523620100, 266912646, -9094956, 86021, ]
      [37479602160, -224877612960, 591525894960, -896251356000, 864242379000, -553115122560, 237742464960, -67926418560, 12486474000, -1387386000, 83243160, -2162160, 13860, ]
                   };

  padeLogm1Coeffs{13} = {
      [0, 1873980108000, -11243880648000, 29615132306760, -45006755593800, 43627479075180, -28152069865920, 12250056650832, -3563611787736, 672585013100, -77730933280, 4958660070, -142985206, 1145993, ]
      [-1873980108000, 12180870702000, -35080907621760, 58955414197680, -64081971954000, 47187633893400, -23968321977600, 8388912692160, -1986847742880, 306612306000, -28857628800, 1475674200, -32792760, 180180, ]
                   };

  padeLogm1Coeffs{14} = {
      [0, 7228208988000, -46983358422000, 135461990664000, -228224006010000, 249142085872680, -184684755315060, 94728347588160, -33621181874280, 8122388490216, -1289496187980, 126500788960, -6892627560, 170388330, -1171733, ]
      [-7228208988000, 50597462916000, -158351319126000, 292340896848000, -353732485186080, 294777070988400, -173021324275800, 71904965932800, -20972281730400, 4194456346080, -551902150800, 44598153600, -1967565600, 37837800, -180180, ]
                   };

  padeLogm1Coeffs{15} = {
      [0, -27949074753600, 195643523275200, -612871808749200, 1133865049917600, -1376941686840720, 1153782781351200, -682621157146500, 286868110042800, -84974134245000, 17363298672720, -2354416335180, 198670814160, -9356498760, 200495280, -1195757, ]
      [27949074753600, -209618060652000, 708364480824000, -1425161872134000, 1900215829512000, -1768662425930400, 1179108283953600, -568498636906200, 197738656315200, -48935324037600, 8388912692160, -953285533200, 66897230400, -2572970400, 43243200, -180180, ]
                   };

  padeLogm1Coeffs{16} = {
      [0, -216605329340400, 1624539970053000, -5494322278645200, 11074238598723300, -14811033908951280, 13849545580710840, -9294080834610000, 4522100920558650, -1592547815086800, 400800618898680, -70300030535280, 8245256756100, -605173290960, 24886594440, -466708208, 2436559, ]
      [216605329340400, -1732842634723200, 6288541819560000, -13695046629264000, 19952266209876000, -20522330958729600, 15328407691396800, -8422202028240000, 3410991821437200, -1010664243388800, 215315425765440, -32030393915520, 3177618444000, -195545750400, 6616209600, -98017920, 360360, ]
                   };

  p = padeLogm1Coeffs{m}{1};
  q = padeLogm1Coeffs{m}{2};

end

% $Id: adimat_logm.m 3970 2013-11-01 15:14:42Z willkomm $
